#'Print method for s2dv_cube objects
#' 
#'This is an S3 method of the generic 'print' for the class 's2dv_cube'. When 
#'you will call 'print' on an 's2dv_cube' object, this method will display the 
#'content of the object in a clear and informative way.
#' 
#'@param x An 's2dv_cube' object
#'@export
print.s2dv_cube <- function(x) {
  cat("'s2dv_cube'\n")
  cat("Data         ", "[" , paste0(x$data[1:5], collapse = ", "), '...', "]", "\n")
  cat("Dimensions   ", "(", paste(names(x$dims), x$dims, sep = " = ", collapse = ', '), ")", "\n")
  cat("Coordinates  \n")
  for (coord in names(x$coords)) {
    cat("  ", coord, ":", paste(x$coords[[coord]], collapse = ", "), "\n")
  }
  cat("Attributes   \n")
  for (attr_name in names(x$attrs)) {
    if (attr_name == "Variable") {
      cat("  ", "varName  :", x$attrs$Variable$varName, "\n")
      cat("  ", "metadata : ", "\n")
      for (metadata_i in names(x$attrs$Variable$metadata)) {
        cat("  ", "  ", metadata_i, "\n")
          for (info_i in names(x$attrs$Variable$metadata[[metadata_i]])) {
            if (info_i == 'units') {
              cat("  ", "  ", "  units :", x$attrs$Variable$metadata[[metadata_i]][[info_i]], "\n")
            } else if (info_i %in% c('longname', 'long_name')) {
              cat("  ", "  ", "  long name :", x$attrs$Variable$metadata[[metadata_i]][[info_i]], "\n")
            }
          }
      }
    } else {
      cat("  ", attr_name, " : ")
      .print_beginning(x$attrs[[attr_name]])
    }
  }
}


## Auxiliary function for the print method
.print_beginning <- function(x, n = 5) {
  if (inherits(x, 'numeric') | inherits(x, 'POSIXct') | inherits(x, 'Date')) {
    if (length(x) <= n) {
      cat(as.character(x), "\n")
    } else {
      cat(paste0(as.character(x[seq_len(n)])), "...", "\n")
    }
  } else if (inherits(x, 'list')) {
    cat("\n")
    for (param in names(x)) {
      cat("      ", "(", param,")", " : ")
      cat(paste0(names(x[[param]]), collapse = ", "), "\n")
    }
  } else {
    if (length(x) > 1) {
      cat(x[[1]], "...", "\n")
    } else {
      cat(x[[1]], "\n")
    }
  }
}
