#*****************************************************************************
# Hands-On exercise: multiApply1
# An-Chi - 30/05/2022
#*****************************************************************************

# 0. Libraries
library(startR)
library(multiApply)

# PART I: [apply]
# 1. Search “apply R” and see which “apply” function we have (e.g., lapply, sapply, mapply, etc.)

## APPLY
arr_1 <- array(1,c(10,5,2))
arr_1
apply(arr_1,3,sum)
## LAPPLY
result <- lapply(arr_1,'-')
result
as.numeric(result)
## SAPPLY
result <- tapply(arr_1,sum)
## TAPPLY

## MAPPLY
mapply(rep, 1:4, 4:1)

# 2. Run the startR script https://earth.bsc.es/gitlab/es/startR/-/blob/master/inst/doc/usecase/ex1_2_exp_obs_attr.R#L23-95 (line 23-95; change retrieve = FALSE to TRUE) to get two data arrays. 
# startR_script.R
#----------------------------------------------------------------------------------------------
# exp
repos_exp <- paste0('/esarchive/exp/ecearth/a1tr/cmorfiles/CMIP/EC-Earth-Consortium/',
                    'EC-Earth3/historical/r24i1p1f1/Amon/$var$/gr/v20190312/',
                    '$var$_Amon_EC-Earth3_historical_r24i1p1f1_gr_$sdate$01-$sdate$12.nc')
exp <- Start(dat = repos_exp,
             var = 'tas',
             sdate = as.character(c(2005:2008)),
             time = indices(1:3),
             lat = 'all',
             lat_reorder = Sort(),
             lon = 'all',
             lon_reorder = CircularSort(0, 360),
             synonims = list(lat = c('lat', 'latitude'),
                             lon = c('lon', 'longitude')),
             return_vars = list(lon = NULL,
                                lat = NULL,
                                time = 'sdate'),
             retrieve = TRUE)
lats <- attr(exp, 'Variables')$common$lat
lons <- attr(exp, 'Variables')$common$lon
## The 'time' attribute is a two-dim array
dates <- attr(exp, 'Variables')$common$time
dim(dates)
dates
repos_obs <- '/esarchive/recon/ecmwf/erainterim/monthly_mean/$var$_f6h/$var$_$date$.nc'
obs <- Start(dat = repos_obs,
             var = 'tas',
             date = unique(format(dates, '%Y%m')),
             time = values(dates),  #dim: [sdate = 4, time = 3]
             lat = 'all',
             lat_reorder = Sort(),
             lon = 'all',
             lon_reorder = CircularSort(0, 360),
             time_across = 'date',
             merge_across_dims = TRUE,
             split_multiselected_dims = TRUE,
             synonims = list(lat = c('lat', 'latitude'),
                             lon = c('lon', 'longitude')),
             return_vars = list(lon = NULL,
                                lat = NULL,
                                time = 'date'),
             retrieve = TRUE)
#----------------------------------------------------------------------------------------------

# 3. Calculate spatial mean of the two arrays using apply(). Check the output dimensions and class.
dim(exp)
dim(obs)
mean_exp <- apply(exp,6,mean)
mean_obs <- apply(obs,6,mean)
mean_exp
class(mean_exp)
class(mean_obs)
dim(mean_exp)
length(mean_obs)

# 4. Reshape the two arrays: (1) Remove the dimensions that have length = 1 by function drop(). 
# (2) Combine the 'sdate' and 'time' dimensions; the dimension should become [time = 12, lat = 256, lon = 512]. Name the two new objects as 'exp2' and 'obs2'.
exp_1 <- drop(exp)
obs_1 <- drop(obs)
dim(exp_1)
dim(obs_1)

# 5. Do point 3 again with the reshaped arrays. Check the output dimensions and class. What is the difference from point 3? [Eva] Nothing??
dim(exp_1)
dim(obs_1)
mean_exp_1 <- apply(exp_1,4,mean)
mean_obs_1 <- apply(obs_1,4,mean)

# --> [Eva] It is possible to put the dimension name in the apply function (and not the number)?
dim(mean_exp_1)
dim(mean_obs_1)
class(mean_exp_1)
length(mean_obs_1)
mean_obs_1

#-----------------------------------------------------------------------------------------------------------------------------------------
# PART I: [multiApply 1]

# 6. Use multiApply::Apply to do the same calculation to exp2 and obs2. Check the dimensions and the class of the outputs.
exp2 <- exp
obs2 <- obs
dim(exp2)
mean_exp2 <- Apply(exp2,6,mean)
length(mean_exp2)
summary(mean_exp2)

mean_obs2 <- Apply(obs2,6,mean)
length(mean_obs2)
summary(mean_obs2)

# 7. Calculate time average of exp2 and obs2 using Apply(). Use the timer below to measure the consuming time. Assign parameter "ncores = 8" in Apply() and see the difference.

print('Start again')
t1 <- Sys.time()
mean_exp2 <- Apply(exp2,6,mean,ncores =8)
mean_obs2 <- Apply(obs2,6,mean,ncores =8)
t2 <- Sys.time()
print(t2 - t1)

# Time consumed ~ 0.40 secs (ncores = 1)
# Time consumed ~ 0.30 secs (ncores = 8)