#*****************************************************************************
# Hands-On exercise: multiApply2
# An-Chi - 30/05/2022
#*****************************************************************************

# 0. Libraries
library(multiApply)
library(ClimProjDiags)
library(s2dv)
library(peakRAM)

#0. Define a function that receives 3 multidimensional arrays:
#  	- One with the dimensions 'time', 'lat' and 'lon'.
#  	- One with the dimensions 'lat' and 'lon'.
#  	- One with the dimensions 'time' and 'lat'.
#   Plus the following additional parameters:
#  	- a Boolean parameter called revert_order
#  	- the '...' parameter

# The function should:
# 	a) Take the second array and subtract it to each time step in the first array.
# 	b) For each time step in the first array, all values along the 'lon' dimension for each 'lat' 
# should be multiplied by the corresponding value in the third array.
# 	c) For each time step, the area average is computed (using the mean() function, to which the 
# additional parameters in '...' are forwarded).
# 	d) If revert_order == TRUE, b) should be performed before a).
# The function should return an array with the dimension 'time'.
#-----------------------------------------------------------------------------------
# (time,lat,lon) = (i,j,k)
# Definition of the function


func_multiapply2 <- function(a1_time_lat_lon, a2_lat_lon, a3_time_lat, revert_order = FALSE, ...) {
    if (revert_order == TRUE) {
        for (i in 1:dim(a1_time_lat_lon)['time']) {
            for (j in 1:dim(a1_time_lat_lon)['lat']) { 
                for (k in 1:dim(a1_time_lat_lon)['lon']) { 
                    a1_time_lat_lon[i,j,k] <- a1_time_lat_lon[i,j,k] * a3_time_lat[i,j] # b)
                }
            }
        }
        for (i in 1:dim(a1_time_lat_lon)['time']) {
            a1_time_lat_lon[i,,] <- a1_time_lat_lon[i,,] - a2_lat_lon[,] # a)
        }
    } else {
    for (i in 1:dim(a1_time_lat_lon)['time']) {
            a1_time_lat_lon[i,,] <- a1_time_lat_lon[i,,] - a2_lat_lon[,] # a)
            for (j in 1:dim(a1_time_lat_lon)['lat']) {
                for (k in 1:dim(a1_time_lat_lon)['lon']) {
                    a1_time_lat_lon[i,j,k] <- a1_time_lat_lon[i,j,k] * a3_time_lat[i,j] # b)
                }
            }
        }
    }
    #print(dim(a1_time_lat_lon)['time'])
    # av_area <- array(1,c(dim(a1_time_lat_lon)['time']))
    # for (i_time in 1:dim(a1_time_lat_lon)['time']) { 
    #     av_area[i_time] <- mean(a1_time_lat_lon[i_time, , ]) 
    # }
    av_area_apply <- apply(a1_time_lat_lon,1,mean)
    return(av_area_apply) # c)
}

#------------------------------------------------------------------------------------------------------
# Exercise 1

# 1. Once you have all this, write a script that:
#  - Creates three sample arrays with the expected dimensions and runs the function by sending the three arrays as parameters

# array declaration
t_dim <- 2
lon_dim <- 5
lat_dim <- 10

dat1_ex1 <- array(1:(t_dim*lat_dim*lon_dim), c(time = t_dim,lat = lat_dim,lon = lon_dim))
dat2_ex1 <- array(1:(lat_dim*lon_dim), c(lat = lat_dim,lon = lon_dim))
dat3_ex1 <- array(1:(t_dim*lat_dim), c(time = t_dim,lat = lat_dim))

dim(dat1_ex1)
#apply(a1_ex1,1,mean)
# use the function as result
result1 <- func_multiapply2(dat1_ex1, dat2_ex1, dat3_ex1)
result1_b <- func_multiapply2(dat1_ex1, dat2_ex1, dat3_ex1, revert_order = TRUE, na.rm = TRUE)

as.vector(result1)
as.vector(result1_b)

# Exercise 2:
# 2. after, write a script that:
# - Creates one array (dat1) with dimensions 'member', 'sdate', 'time', 'lat', 'lon', 'depth' with random values (no matter the length of the dimensions)
# 	- Creates one array (dat2) with dimensions 'lat' and 'lon' with random values (same length as in the first array)
# 	- Creates one array (dat3) with dimensions 'member', 'time', 'lat', 'depth' with random values (same length as in the first array)
# 	- Applies the function to the arrays using multiApply, in a way that the function is applied to each possible combination of sub arrays of dimensions
# ('time', 'lat', 'lon'), ('lat', 'lon') and ('time', 'lat'), resulting in a array of dimensions ‘time’ ‘member’ ‘sdate’ ‘depth’


m_dim <- 3
s_dim <- 3
depth_dim <- 4

length_a1l_dims <- m_dim*s_dim*t_dim*lon_dim*lat_dim*depth_dim

dat1_ex2 <- array(rnorm(length_a1l_dims), c(member = m_dim, sdate = s_dim, time = t_dim, lat = lat_dim, lon = lon_dim, depth = depth_dim))
dat2_ex2 <- array(rnorm(lat_dim*lon_dim), c(lat = lat_dim, lon = lon_dim))
dat3_ex2 <- array(rnorm(m_dim*t_dim*lat_dim*depth_dim), c(member = m_dim, time = t_dim, lat = lat_dim, depth = depth_dim))

dim(dat1_ex2)
dim(dat2_ex2)
dim(dat3_ex2)


#Using different arrays
result3 <- Apply(list(dat1_ex2,dat2_ex2,dat3_ex2), target_dims = list(c('time', 'lat', 'lon'),c('lat', 'lon'),c('time', 'lat')),func_multiapply2, output_dims = 'time')
dim(result3[[1]])

# - Apply the function to the arrays again but send the parameter 'revert_order = TRUE' to the atomic function, and the parameter 
# 'na.rm = TRUE' to the mean() function inside the atomic function

result3_b <- Apply(list(dat1_ex2,dat2_ex2,dat3_ex2), target_dims = list(c('time', 'lat', 'lon'),c('lat', 'lon'),c('time', 'lat')),func_multiapply2, 
                revert_order = TRUE, na.rm = TRUE, output_dims = 'time')
dim(result3_b[[1]])

#---------------------------------------------------------------------------------------------------------------

# Exercise 3
# 3. [Reorder arrays]
# - Randomly change the order of the dimensions of the first array (dat1) with aperm(), preserving the original arrays --> OK
# - Change the order of the dimensions of the created arrays with s2dv:::.aperm2(), preserving the original arrays --> OK (hidden function)
# - Change the order of the dimensions of the created arrays with s2dv::Reorder, preserving the original arrays
#     → What is the difference between these? --> Computational time is similar

t1 <- Sys.time()
dat1_ex3 <- dat1_ex2
dim(dat1_ex3)
dat1_ex3 <- aperm(dat1_ex3,sample.int(length(dim(dat1_ex3))))
dim(dat1_ex3)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 0.08729243 secs

t1 <- Sys.time()
dat2_ex3 <- dat1_ex2
dim(dat2_ex3)
dat2_ex3 <- s2dv:::.aperm2(dat2_ex3,sample.int(length(dim(dat2_ex3))))
dim(dat2_ex3)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 0.04052567 secs


t1 <- Sys.time()
dat3_ex3 <- dat1_ex2
dim(dat3_ex3)
dat3_ex3 <- s2dv::Reorder(dat3_ex3,sample.int(length(dim(dat3_ex3))))
dim(dat3_ex3)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 0.0642519 secs

# - Apply the function to the array returned by s2dv::Reorder with dat2 and dat3. Did it work? Is the result identical? 
# Check the dimensions and values.
dat4_ex3 <- dat2_ex2
dat5_ex3 <- dat3_ex2

dim(dat4_ex3)
dim(dat5_ex3)
dat4_ex3 <- s2dv::Reorder(dat4_ex3,sample.int(length(dim(dat4_ex3))))
dat5_ex3 <- s2dv::Reorder(dat5_ex3,sample.int(length(dim(dat5_ex3))))
dim(dat4_ex3)
dim(dat5_ex3)

# Exercise 5
# [efficiency test]
# *To see the results more clearly, enlarge the data size (but don’t crash the machine)
# 	- Compare the wall clock time spent in the apply before reordering the arrays and the time spent in the apply after reordering the arrays
# 	- Compare the peak RAM memory consumption of both executions using the peakRAM package



m_dim <- 4
s_dim <- 30
t_dim <- 10
lat_dim <- 100
lon_dim <- 100
depth_dim <- 4
length_a1l_dims <- m_dim*s_dim*t_dim*lon_dim*lat_dim*depth_dim

dat1_ex5 <- array(rnorm(length_a1l_dims), c(member = m_dim, sdate = s_dim, time = t_dim, lat = lat_dim, lon = lon_dim, depth = depth_dim))

length(dat1_ex5)
dim(dat1_ex5)

t1 <- Sys.time()
dat2_ex5 <- dat1_ex5
dim(dat2_ex5)
dat2_ex5 <- aperm(dat2_ex5,sample.int(length(dim(dat2_ex5))))
dim(dat2_ex5)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 1.592304 secs

t1 <- Sys.time()
dat2_ex5 <- dat1_ex5
dim(dat2_ex5)
dat2_ex5 <- s2dv:::.aperm2(dat2_ex5,sample.int(length(dim(dat2_ex5))))
dim(dat2_ex5)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 1.541103 secs

t1 <- Sys.time()
dat3_ex5 <- dat1_ex5
dim(dat3_ex5)
dat3_ex5 <- s2dv::Reorder(dat3_ex5,sample.int(length(dim(dat3_ex5))))
dim(dat3_ex5)
t2 <- Sys.time()
print(t2-t1)
# Time difference of 2.806826 secs


peakRAM(aperm(dat1_ex5,sample.int(length(dim(dat1_ex5)))))
# Function_Call Elapsed_Time_sec
# 1 aperm(dat1_ex5,sample.int(length(dim(dat1_ex5))))             1.68
#   Total_RAM_Used_MiB Peak_RAM_Used_MiB
# 1              366.2             366.2

peakRAM(s2dv:::.aperm2(dat1_ex5,sample.int(length(dim(dat1_ex5)))))
# Function_Call Elapsed_Time_sec
# 1 s2dv:::.aperm2(dat1_ex5,sample.int(length(dim(dat1_ex5))))            0.619
#   Total_RAM_Used_MiB Peak_RAM_Used_MiB
# 1              366.3             366.3

peakRAM(s2dv::Reorder(dat1_ex5,sample.int(length(dim(dat1_ex5)))))
# Function_Call Elapsed_Time_sec
# 1 s2dv::Reorder(dat1_ex5,sample.int(length(dim(dat1_ex5))))            1.074
#   Total_RAM_Used_MiB Peak_RAM_Used_MiB
# 1              366.3             366.3



# 	- Remove the dimension names of the three arrays and manage to apply the function with multiApply using numeric indices.

dat1_ex6 <- dat1_ex2
dat2_ex6 <- dat2_ex2
dat3_ex6 <- dat3_ex2


dim(dat1_ex6) <- unname(dim(dat1_ex6))
dim(dat1_ex6)

# dim(dat2_ex6) <- unname(dim(dat2_ex6))
# dim(dat3_ex6) <- unname(dim(dat3_ex6))

names(dim(dat1_ex6))
# dim(dat2_ex6)
# dim(dat3_ex6)
seq(length(dim(dat1_ex6)))

dim(dat1_ex6)
length(dim(dat1_ex6))

#dat1_ex6 <- Apply(dat1_ex6, target_dims = names(dim(dat1_ex6)),Reorder,sample.int(length(dim(dat1_ex6))))
dim(dat1_ex6)
dat4_ex6 <- Apply(dat1_ex6, target_dims = seq(length(dim(dat1_ex6))),aperm)
target_dims = dim(dat)
# it will be faster to use indices, not names

dim(dat4_ex6)
dat4_ex6
#dat5_ex6 <- Apply(dat4_ex6, target_dims = seq(1:length(dim(dat4_ex6))),Reorder)


# 	- Use package “profvis” in RStudio to see the memory and time efficiency of both executions
library(profvis)
?profvis
profvis({dat2_ex5 <- aperm(dat2_ex5,sample.int(length(dim(dat2_ex5))))
})
profvis({
dat2_ex5 <- s2dv:::.aperm2(dat2_ex5,sample.int(length(dim(dat2_ex5))))
})
profvis({dat2_ex5 <- s2dv::Reorder(dat2_ex5,sample.int(length(dim(dat2_ex5))))
})

m_dim <- 3
s_dim <- 3
depth_dim <- 4

length_all_dims <- m_dim*s_dim*t_dim*lon_dim*lat_dim*depth_dim

dat1_ex7 <- array(rnorm(length_all_dims), c(member = m_dim, sdate = s_dim, time = t_dim, lat = lat_dim, lon = lon_dim, depth = depth_dim))
dat2_ex7 <- array(rnorm(lat_dim*lon_dim), c(lat = lat_dim, lon = lon_dim))
dat3_ex7 <- array(rnorm(m_dim*t_dim*lat_dim*depth_dim), c(member = m_dim, time = t_dim, lat = lat_dim, depth = depth_dim))

profvis({result3_b <- Apply(list(dat1_ex7,dat2_ex7,dat3_ex7), target_dims = list(c('time', 'lat', 'lon'),c('lat', 'lon'),c('time', 'lat')),func_multiapply2, 
                            revert_order = FALSE, na.rm = TRUE, output_dims = 'time')})

?benchmark
# Efficiency tests: 1) clock 2) peakRAM (only shows the peak memory - simple) 3) profvis: 4) rbenchmark 
# IMPORTANT EFFICIENCY
# STARTR