#-------------------------------------------------------------------------------
# Step (0): log in into autosubmit machine (ssh bscesautosubmit01)
# Step (1): Create a new autosubmit experiment
#  module load autosubmit/3.14.0-foss-2015a-Python-2.7.9
#  autosubmit expid -H nord3v2 -d <Description> # Write anything you like in the description, like "Auto-S2S test" for example.
  # Test: a6mb
# Step (2): Modify the recipe in Run section
# Step (3): Source MODULES
# Step (4): Rscript split.R /esarchive/scratch/erifarov/rpackages/sunset/test-autosubmit/autosubmit-wf/reciepe-autosubmit.yml
#-------------------------------------------------------------------------------
# Load modules
source("modules/Loading/Loading.R")
source("modules/Calibration/Calibration.R")
source("modules/Anomalies/Anomalies.R")
source("modules/Skill/Skill.R")
source("modules/Saving/Saving.R")
source("modules/Visualization/Visualization.R")

#### -- autosubmit
# args = commandArgs(trailingOnly = TRUE)
# recipe_file <- args[1]
# recipe <- read_atomic_recipe(recipe_file)
#### -- interactive
recipe_file <- "/esarchive/scratch/erifarov/rpackages/sunset/dev-scorecards/recipe_scorecards.yml"

# source dev files
filepath <- "/esarchive/scratch/erifarov/rpackages/sunset/dev-scorecards/modified_files_sunset/"
f <- list.files(filepath)  
ff <- lapply(f, function(x) paste0(filepath, x))  
invisible(lapply(ff,source))

recipe <- prepare_outputs(recipe_file)


# Load datasets
data <- Loading(recipe)
# Calibrate datasets
data <- Calibration(recipe, data)
# Compute anomalies
data <- Anomalies(recipe, data)
# Compute skill metrics 
skill_metrics <- Skill(recipe, data)
# Compute percentiles and probability bins
probabilities <- Probabilities(recipe, data)
# Plot data
Visualization(recipe, data, skill_metrics, probabilities,
	      significance = T)
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
> metrics
 [1] "rps"             "rpss"            "crps"            "crpss"          
 [5] "frpss"           "bss10"           "bss90"           "mean_bias"      
 [9] "mean_bias_ss"    "rps_clim_syear"  "rps_syear"       "crps_clim_syear"
[13] "crps_syear" 

syear = '24'
summary(skill_metrics[['rps']])

rps_syear <- array(rnorm(6*5*5*24, 0.5351, 0.082), dim = c(dim(skill_metrics[['rps']]), syear = 24))
rps_clim_syear <- array(rnorm(6*5*5*24, 0.4478, 0.082), dim = c(dim(skill_metrics[['rps_clim']]), syear = 24))

crps_clim_syear
crps_syear

skill_metrics[['rps_syear']] <- rps_syear
skill_metrics[['rps_clim_syear']] <- rps_clim_syear

save_metrics(recipe = recipe, skill = skill_metrics, data_cube = data$hcst, agg = 'global')

#-------------------------------------------------------------------------------

# source("modules/Loading/Loading.R")
# source("modules/Calibration/Calibration.R")
# source("modules/Anomalies/Anomalies.R")
# source("modules/Skill/Skill.R")
# source("modules/Saving/Saving.R")

# args = commandArgs(trailingOnly = TRUE)
# recipe_file <- args[1]
# recipe <- read_atomic_recipe(recipe_file)
# # Load datasets
# data <- load_datasets(recipe)
# # Compute anomalies
# data <- compute_anomalies(recipe, data)
# # Compute skill metrics 
# skill_metrics <- compute_skill_metrics(recipe, data)
